from typing import List

import numpy as np
import pytest
from numpy.testing import assert_equal
from sklearn.tree._tree import Tree

from coniferest.coniferest import Coniferest


class ConiferestImpl(Coniferest):
    def fit(self, data, labels=None):
        super().fit(data, labels)

    def fit_known(self, data, known_data=None, known_labels=None):
        super().fit_known(data, known_data, known_labels)

    def score_samples(self, samples):
        return super().score_samples(samples)

    def feature_signature(self, x):
        raise NotImplementedError()

    def feature_importance(self, x):
        raise NotImplementedError()


def build_one_tree(random_seed) -> Tree:
    shape = 256, 16

    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal(shape)

    coniferest = ConiferestImpl(trees=None, n_subsamples=data.shape[0], max_depth=None, random_seed=random_seed)
    return coniferest.build_one_tree(data)


def assert_tree_equal(a, b):
    assert_equal(a.children_left, b.children_left)
    assert_equal(a.children_right, b.children_right)
    assert_equal(a.value, b.value)
    assert_equal(a.feature, b.feature)
    assert_equal(a.threshold, b.threshold)


def test_reproducibility_build_one_tree():
    """
    Are we able to reproduce tree building?
    """
    random_seed = np.random.randint(1 << 16)
    assert_tree_equal(build_one_tree(random_seed), build_one_tree(random_seed))


@pytest.mark.regression
def test_regression_build_one(regression_data):
    tree = build_one_tree(0)
    regression_data.check_with(assert_tree_equal, tree)


def build_trees(random_seed) -> List[Tree]:
    n_trees = 8
    n_subsamples = 256
    shape = n_subsamples * n_trees, 16

    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal(shape)

    coniferest = ConiferestImpl(trees=None, n_subsamples=n_subsamples, max_depth=None, random_seed=random_seed)
    return coniferest.build_trees(data, n_trees)


def test_reproducibility_build_trees():
    """
    Are we able to reproduce Coniferest.build_trees
    """
    random_seed = np.random.randint(1 << 16)

    trees1 = build_trees(random_seed)
    trees2 = build_trees(random_seed)

    for tree1, tree2 in zip(trees1, trees2):
        assert_tree_equal(tree1, tree2)


@pytest.mark.regression
def test_regression_build_trees(regression_data):
    trees = build_trees(0)
    regression_data.check_with(
        lambda actual, desired: [assert_tree_equal(a, b) for a, b in zip(actual, desired)],
        trees,
    )



import numpy as np
import pytest

from coniferest.aadforest import AADForest
from coniferest.datasets import non_anomalous_outliers, ztf_m31
from coniferest.label import Label
from coniferest.pineforest import PineForest
from coniferest.session import Session


@pytest.mark.e2e
@pytest.mark.regression
def test_e2e_ztf_m31():
    """Basically the same example as in the docs"""

    class Callback:
        """Say NO for first few objects, then say YES and terminate"""

        counter = 0

        def __init__(self, n_iter):
            self.n_iter = n_iter

        def decision(self, _metadata, _data, _session) -> Label:
            self.counter += 1
            if self.counter < self.n_iter:
                return Label.REGULAR
            return Label.ANOMALY

        def on_decision(self, _metadata, _data, session) -> None:
            if self.counter >= self.n_iter:
                session.terminate()

    callback = Callback(2)

    data, metadata = ztf_m31()
    model = AADForest(
        n_trees=128,
        n_subsamples=256,
        random_seed=0,
    )
    session = Session(
        data=data,
        metadata=metadata,
        model=model,
        decision_callback=callback.decision,
        on_decision_callbacks=[callback.on_decision],
    )
    session.run()

    assert len(session.known_labels) == callback.n_iter

    oid = 695211200075348
    idx = np.where(metadata == oid)[0][0]
    assert idx in session.known_labels
    assert session.known_labels[idx] == Label.ANOMALY


# We mark it long because for some platforms it sometimes takes an hour to run it for the AAD case
# https://github.com/snad-space/coniferest/issues/121
@pytest.mark.long
@pytest.mark.e2e
@pytest.mark.regression
@pytest.mark.parametrize(
    "model,n_iter,last_idx",
    [
        (AADForest(n_trees=128, random_seed=0), 48, 1117),
        (PineForest(n_trees=128, n_spare_trees=512, random_seed=0), 34, 1109),
    ],
)
def test_non_anomalous_outliers(model, n_iter, last_idx):
    # Number of normal objects
    n_regular = 1 << 10
    # Number of outliers in each of three groups, only one of which is anomalous
    n_anomalies = 1 << 5
    data, metadata = non_anomalous_outliers(inliers=n_regular, outliers=n_anomalies)

    def terminate_after_all_detected(_metadata, _data, session):
        if session.known_anomalies.size == n_anomalies:
            session.terminate()

    session = Session(
        data=data,
        metadata=metadata,
        model=model,
        decision_callback=lambda metadata, _data, _session: metadata,
        on_decision_callbacks=[terminate_after_all_detected],
    )
    session.run()

    assert len(session.known_labels) == n_iter
    assert session.current == last_idx



import numpy as np
import pytest

from coniferest.aadforest import AADForest
from coniferest.datasets import single_outlier
from coniferest.label import Label


def test_scores_negative():
    data = np.arange(1024.0).reshape(256, 4)
    forest = AADForest(n_trees=10, random_seed=0).fit(data)
    scores = forest.score_samples(data)
    assert np.all(scores < 0)


def test_single_outlier():
    data, _metadata = single_outlier()
    forest = AADForest(n_trees=10, random_seed=0).fit(data)
    scores = forest.score_samples(data)
    # Outlier goes last and must have the lowest score
    assert np.argmin(scores) == data.shape[0] - 1


def test_prior_influence_value():
    data, _metadata = single_outlier()
    forest = AADForest(n_trees=10, random_seed=0, prior_influence=2.0).fit(data)
    scores = forest.score_samples(data)
    # Outlier goes last and must have the lowest score
    assert np.argmin(scores) == data.shape[0] - 1


def test_prior_influence_callable():
    data, _metadata = single_outlier()
    forest = AADForest(
        n_trees=10,
        random_seed=0,
        prior_influence=lambda ac, nc: np.max(1.0, 0.5 / (ac + nc)),
    ).fit(data)
    scores = forest.score_samples(data)
    # Outlier goes last and must have the lowest score
    assert np.argmin(scores) == data.shape[0] - 1


@pytest.mark.regression
def test_regression_fit_known(regression_data):
    random_seed = 0
    n_samples = 1024
    n_features = 16
    n_known = 16
    n_trees = 128
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    known_data = data[rng.choice(n_samples, n_known, replace=False)]
    known_labels = rng.choice([-1, 1], n_known, replace=True)

    forest = AADForest(n_trees=n_trees, random_seed=random_seed)
    forest.fit(data)
    pre_fit_known_scores = forest.score_samples(data)

    forest.fit_known(data, known_data=known_data, known_labels=known_labels)
    scores = forest.score_samples(data)

    assert not np.allclose(pre_fit_known_scores, scores, rtol=1e-3), "Scores must change after fit_known"

    regression_data.assert_allclose(scores)


@pytest.mark.benchmark
@pytest.mark.long
def test_benchmark_fit_known(n_jobs, benchmark):
    benchmark.group = f"AADForest.fit_known {n_jobs = :2d}"
    benchmark.name = "coniferest.aadforest.AADForest"

    random_seed = 0
    n_samples = 1024
    n_features = 16
    n_known = 16
    n_trees = 128
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    known_data = data[rng.choice(n_samples, n_known, replace=False)]
    known_labels = rng.choice([-1, 1], n_known, replace=True)

    forest = AADForest(n_trees=n_trees, n_jobs=n_jobs, random_seed=random_seed)
    forest.fit(data)

    benchmark(forest.fit_known, data, known_data=known_data, known_labels=known_labels)


@pytest.mark.benchmark
@pytest.mark.long
@pytest.mark.parametrize("n_samples", [1 << 10, 1 << 16])
@pytest.mark.parametrize("n_trees", [1 << 10, 1 << 14])
def test_benchmark_loss_gradient(n_samples, n_trees, n_jobs, benchmark):
    benchmark.group = f"AADEvaluator.loss_graident {n_samples = :6d} {n_trees = :4d}, {n_jobs = :2d}"
    benchmark.name = "coniferest.aadforest.AADEvaluator"

    random_seed = 0
    n_features = 16
    n_known = 16
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    known_data = data[rng.choice(n_samples, n_known, replace=False)]
    known_labels = rng.choice([-1, 1], n_known, replace=True)

    forest = AADForest(n_trees=n_trees, n_jobs=n_jobs, random_seed=random_seed)
    forest.fit(data)

    anomaly_count = np.count_nonzero(known_labels == Label.ANOMALY)
    nominal_count = np.count_nonzero(known_labels == Label.REGULAR)
    scores = forest.score_samples(data)
    q_tau = np.quantile(scores, 1.0 - forest.tau)

    benchmark(
        forest.evaluator.loss_gradient,
        forest.evaluator.weights,
        known_data,
        known_labels,
        anomaly_count,
        nominal_count,
        q_tau,
    )



cimport numpy as np

# Selector is the representation of decision tree nodes:
# either branches or leafs.
cdef packed struct selector_t:
    # Feature index to branch on
    np.int32_t feature
    # Index of left subtree
    np.int32_t left
    # Feature value to branch on
    np.float64_t value
    # Index of right subtree
    np.int32_t right
    # Log of samples count covered by this node
    np.float32_t log_n_node_samples

    # In case of leaf the fields have special values:
    # * feature == -1
    # * left == leaf_id
    # * value == resulting decision score
    # * right == -1


# For bitness polymorphism
ctypedef fused floating:
    np.float32_t
    np.float64_t


# Calculate paths sum on data based on trees,
# that are located by indeces in selectors list
cdef void _paths_sum(selector_t [::1] selectors,
                     np.int64_t [::1] indices,
                     floating [:, ::1] data,
                     np.float64_t [::1] paths,
                     np.float64_t [::1] weights=*,
                     int num_threads=*)

cdef void _paths_sum_transpose(selector_t [::1] selectors,
                               np.int64_t [::1] indices,
                               floating [:, ::1] data,
                               np.float64_t [::1] values,
                               np.float64_t [::1] weights=*,
                               int num_threads=*)

cdef void _feature_delta_sum(selector_t [::1] selectors,
                             np.int64_t [::1] indices,
                             floating [:, ::1] data,
                             np.float64_t [:, ::1] delta_sum,
                             np.int64_t [:, ::1] hit_count,
                             int num_threads=*)

cdef void _apply(selector_t [::1] selectors,
                 np.int64_t [::1] indices,
                 floating [:, ::1] data,
                 np.int64_t [:, ::1] leafs,
                 int num_threads=*)



import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal
from sklearn.ensemble import IsolationForest as SkIsolationForest

from coniferest.datasets import MalanchevDataset
from coniferest.isoforest import IsolationForest
from coniferest.sklearn.isoforest import IsolationForestEvaluator


@pytest.fixture()
def isoforest_results():
    return IsoforestResults()


class IsoforestResults:
    def __init__(self):
        seed = 622341
        self.dataset = MalanchevDataset(inliers=1000, outliers=50, regions=[1, 1, -1], rng=seed)

        data = self.dataset.data
        trees = 1000

        forest = IsolationForest(n_trees=trees, random_seed=seed + 1)
        self.scores = self.calc_forest_scores(forest, data)
        self.forest = forest

        forest = SkIsolationForest(n_estimators=trees, random_state=seed + 2)
        self.skores0 = self.calc_forest_scores(forest, data)

        forest = SkIsolationForest(n_estimators=trees, random_state=seed + 3)
        self.skores1 = self.calc_forest_scores(forest, data)
        self.skores1_by_evaluator = IsolationForestEvaluator(forest).score_samples(data)

    @staticmethod
    def calc_forest_scores(forest, data):
        forest.fit(data)
        return forest.score_samples(data)


def test_sklearn_isolation_forest_evaluator(isoforest_results):
    """
    Does evaluator scores coinside with the ones computed by sklearn?
    """
    r = isoforest_results
    assert_allclose(
        r.skores1_by_evaluator,
        r.skores1,
        atol=1e-10,
        rtol=0,
        err_msg="sklearn and our results nust be the same",
    )


def test_isolation_forest(isoforest_results):
    """
    Does assigned scores by our isoforest are somewhere near assigned by sklearn's isoforest?
    """
    r = isoforest_results
    diff_sk_to_sk = np.max(np.abs(r.skores0 - r.skores1))
    diff_coni_to_sk = np.max(np.abs(r.skores0 - r.scores))
    assert diff_coni_to_sk < 1.5 * diff_sk_to_sk


def test_serialization(isoforest_results):
    """
    Does (de)serialization work correctly?
    """
    import pickle

    r = isoforest_results
    s = pickle.dumps(r.forest)
    reforest = pickle.loads(s)
    assert_allclose(reforest.score_samples(r.dataset.data), r.scores, atol=1e-12)


def forest_n_features(forest: IsolationForest):
    return forest.evaluator.selectors[0, 0].n_features


def assert_forest_scores(forest1: IsolationForest, forest2: IsolationForest, data=None, n_features=None):
    if data is None:
        if n_features is None:
            raise ValueError("Either data or n_features")
        data = np.random.standard_normal((1024, n_features))
    assert_equal(forest1.score_samples(data), forest2.score_samples(data))


def build_forest(n_features: int, random_seed: int) -> IsolationForest:
    n_trees = 100
    n_subsamples = 256

    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_trees * n_subsamples, n_features))

    forest = IsolationForest(
        n_trees=n_trees,
        n_subsamples=n_subsamples,
        max_depth=None,
        random_seed=random_seed,
    )
    forest.fit(data)
    return forest


def test_reproducibility():
    n_features = 16
    random_seed = np.random.randint(1 << 16)
    forest1 = build_forest(n_features, random_seed)
    forest2 = build_forest(n_features, random_seed)
    assert_forest_scores(forest1, forest2, n_features=n_features)


def test_apply():
    n_features = 16
    n_trees = 100
    n_subsamples = 256

    random_seed = np.random.randint(1 << 16)
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_trees * n_subsamples, n_features))

    forest = IsolationForest(
        n_trees=n_trees,
        n_subsamples=n_subsamples,
        max_depth=None,
        random_seed=random_seed,
    )

    forest.fit(data)

    leafs = forest.apply(data)
    selectors = forest.evaluator.selectors
    leaf_mask = selectors["feature"] < 0
    scores = np.sum(selectors[leaf_mask][leafs]["value"], axis=1)
    scores = -(2 ** (-scores / (forest.evaluator.average_path_length(n_subsamples) * n_trees)))
    assert_allclose(forest.score_samples(data), scores)


@pytest.mark.regression
def test_regression(regression_data):
    random_seed = 0
    n_features = 16
    n_samples = 128
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = build_forest(n_features=n_features, random_seed=random_seed)
    scores = forest.score_samples(data)
    regression_data.assert_allclose(scores)


@pytest.mark.regression
def test_regression_signatures(regression_data):
    random_seed = 0
    n_features = 16
    n_samples = 128
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = build_forest(n_features=n_features, random_seed=random_seed)
    signatures = forest.feature_signature(data)
    regression_data.assert_allclose(signatures)


@pytest.mark.regression
def test_regression_importance(regression_data):
    random_seed = 0
    n_features = 16
    n_samples = 128
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = build_forest(n_features=n_features, random_seed=random_seed)
    importance = forest.feature_importance(data)
    regression_data.assert_allclose(importance)


def test_n_jobs():
    random_seed = 0
    n_features = 16
    n_samples = 1024
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))

    reference_forest = IsolationForest(n_trees=5, random_seed=random_seed)
    reference_forest.fit(data)

    for n_jobs in [1, 2, -1, None]:
        forest = IsolationForest(n_trees=5, n_jobs=n_jobs, random_seed=random_seed)
        forest.fit(data)
        assert_forest_scores(reference_forest, forest, data=data)


@pytest.mark.benchmark
@pytest.mark.long
@pytest.mark.parametrize("n_trees", [128, 1024])
def test_benchmark_fit(n_trees, n_jobs, benchmark):
    benchmark.group = f"IsolationForest.fit {n_trees = :4d}, {n_jobs = :2d}"
    benchmark.name = "coniferest.isoforest.IsolationForest"

    random_seed = 0
    n_samples = 16_384
    n_features = 16
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = IsolationForest(n_trees=n_trees, n_jobs=n_jobs, random_seed=random_seed)

    benchmark(forest.fit, data)


# We need to merge it with previous one when we make interface consistent with sklearn's
# https://github.com/snad-space/coniferest/issues/113
@pytest.mark.benchmark
@pytest.mark.long
@pytest.mark.parametrize("n_trees", [128, 1024])
def test_benchmark_fit_sklearn(n_trees, n_jobs, benchmark):
    benchmark.group = f"IsolationForest.fit {n_trees = :4d}, {n_jobs = :2d}"
    benchmark.name = "sklearn.ensemble.IsolationForest"

    random_seed = 0
    n_samples = 16_384
    n_features = 16
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = SkIsolationForest(n_estimators=n_trees, n_jobs=n_jobs, random_state=random_seed)

    benchmark(forest.fit, data)


@pytest.mark.benchmark
@pytest.mark.long
@pytest.mark.parametrize("n_samples", [1 << 10, 1 << 20])
def test_benchmark_score(n_samples, n_jobs, benchmark):
    benchmark.group = f"IsolationForest.score_samples {n_samples = :7d}, {n_jobs = :2d}"
    benchmark.name = "coniferest.isoforest.IsolationForest"

    random_seed = 0
    n_features = 16
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = IsolationForest(n_trees=128, n_jobs=n_jobs, random_seed=random_seed)
    forest.fit(data)

    benchmark(forest.score_samples, data)


# We need to merge it with previous one when we make interface consistent with sklearn's
# https://github.com/snad-space/coniferest/issues/113
@pytest.mark.benchmark
@pytest.mark.long
@pytest.mark.parametrize("n_samples", [1 << 10, 1 << 20])
def test_benchmark_score_sklearn(n_samples, n_jobs, benchmark):
    benchmark.group = f"IsolationForest.score_samples {n_samples = :7d}, {n_jobs = :2d}"
    benchmark.name = "sklearn.ensemble.IsolationForest"

    random_seed = 0
    n_features = 16
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = SkIsolationForest(n_estimators=128, n_jobs=n_jobs, random_state=random_seed)
    forest.fit(data)

    benchmark(forest.score_samples, data)


@pytest.mark.benchmark
@pytest.mark.long
@pytest.mark.parametrize("n_features", [2, 128])
def test_benchmark_feature_signature(n_features, n_jobs, benchmark):
    benchmark.group = f"IsolationForest.feature_signature {n_features = :3d}, {n_jobs = :2d}"
    benchmark.name = "coniferest.isoforest.IsolationForest"

    random_seed = 0
    n_samples = 1 << 14
    n_trees = 1024
    rng = np.random.default_rng(random_seed)
    data = rng.standard_normal((n_samples, n_features))
    forest = IsolationForest(n_trees=n_trees, n_jobs=n_jobs, random_seed=random_seed)
    forest.fit(data)

    benchmark(forest.feature_signature, data[:1])



import numpy as np
import onnx
import onnxruntime as rt

from coniferest.aadforest import AADForest
from coniferest.datasets import single_outlier
from coniferest.onnx import to_onnx


def _evaluate_onnx(o, data):
    sess = rt.InferenceSession(o.SerializeToString())
    input_name = sess.get_inputs()[0].name
    label_name = "score"
    return sess.run([label_name], {input_name: data})[0].reshape(-1)


def test_onnx_aadforest():
    data, _metadata = single_outlier()
    model = AADForest(n_trees=10, random_seed=0).fit(data)
    scores_model = model.score_samples(data)

    o = to_onnx(model)
    onnx.checker.check_model(o)
    scores_onnx = _evaluate_onnx(o, data.astype(np.float32))

    np.testing.assert_allclose(scores_onnx, scores_model, 1e-5)



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from coniferest.datasets import dev_net_dataset, plasticc_gp, ztf_m31


def test_ztf_m31():
    data, metadata = ztf_m31()

    assert data.shape == (57546, 42)
    assert data.dtype == np.float32
    assert_allclose(data.sum(axis=1)[13], 3.4734121e09, atol=1e-6)

    assert metadata.shape == (data.shape[0],)
    assert metadata.dtype == np.uint64
    assert_allclose(data.mean(), 100942920)


def test_dev_net():
    data, metadata = dev_net_dataset("thyroid")

    assert data.shape == (7200, 21)

    assert metadata.shape == (data.shape[0],)
    assert_array_equal(np.unique(metadata), [-1, 1])


def test_plasticc_gp():
    data, metadata = plasticc_gp()

    assert data.shape == (7204, 374)
    assert data.dtype == np.float32
    assert_allclose(data.sum(axis=1)[13], 70.540085, atol=1e-6)

    assert metadata.shape == (data.shape[0],)
    assert_array_equal(np.unique(metadata), [-1, 1])



import joblib
import numpy as np

from .calc_paths_sum import calc_apply, calc_feature_delta_sum, calc_paths_sum  # noqa
from .utils import average_path_length

__all__ = ["ForestEvaluator"]


class ForestEvaluator:
    selector_dtype = np.dtype(
        [
            ("feature", np.int32),
            ("left", np.int32),
            ("value", np.double),
            ("right", np.int32),
            ("log_n_node_samples", np.float32),
        ]
    )

    def __init__(self, samples, selectors, indices, leaf_count, *, num_threads):
        """
        Base class for the forest evaluators. Does the trivial job:
        * runs calc_paths_sum written in cython,
        * helps to combine several trees' representations into two arrays.

        Parameters
        ----------
        samples
            Number of samples in every tree.

        selectors
            Array with all the nodes of all the trees.

        indices
            Indices of starting and ending nodes of every tree. For example
            for two trees of length `len1` and `len2` it would be:
            [0, len1, len1 + len2]

        leaf_count
            Total number of leafs in all the trees.

        num_threads : int or None
            Number of threads to use for calculations. If None then
        """
        self.samples = samples

        self.selectors = selectors
        self.indices = indices
        self.leaf_count = leaf_count

        if num_threads is None or num_threads < 1:
            # Count of available CPUs is not a simple thing, see loky's implementation here:
            # https://github.com/joblib/joblib/blob/476ff8e62b221fc5816bad9b55dec8883d4f157c/joblib/externals/loky/backend/context.py#L83
            # We ignore OMP_NUM_THREADS for now
            self.num_threads = joblib.cpu_count()
        else:
            self.num_threads = num_threads

    @classmethod
    def combine_selectors(cls, selectors_list):
        """
        Combine several node arrays into one array of nodes and one array of
        start indices.

        Parameters
        ----------
        selectors_list
            List of node arrays to combine.

        Returns
        -------
        Pair of two arrays: node array and array of starting indices.
        """
        lens = [len(sels) for sels in selectors_list]
        full_len = sum(lens)

        selectors = np.empty((full_len,), dtype=cls.selector_dtype)

        indices = np.empty((len(selectors_list) + 1,), dtype=np.int64)
        indices[0] = 0
        indices[1:] = np.add.accumulate(lens)

        for i in range(len(selectors_list)):
            selectors[indices[i] : indices[i + 1]] = selectors_list[i]

        # Assign a unique sequential index to every leaf
        # The index is used for weighted scores
        leaf_mask = selectors["feature"] < 0
        leaf_count = np.count_nonzero(leaf_mask)
        selectors["left"][leaf_mask] = np.arange(0, leaf_count)

        return selectors, indices, leaf_count

    def score_samples(self, x):
        """
        Perform the computations.

        Parameters
        ----------
        x
            Features to calculate scores of. Should be C-contiguous for performance.

        Returns
        -------
        Array of scores.
        """
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        trees = self.indices.shape[0] - 1

        return -(
            2
            ** (
                -calc_paths_sum(self.selectors, self.indices, x, num_threads=self.num_threads)
                / (self.average_path_length(self.samples) * trees)
            )
        )

    def _feature_delta_sum(self, x):
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        return calc_feature_delta_sum(self.selectors, self.indices, x, num_threads=self.num_threads)

    def feature_signature(self, x):
        delta_sum, hit_count = self._feature_delta_sum(x)

        return delta_sum / hit_count / self.average_path_length(self.samples)

    def feature_importance(self, x):
        delta_sum, hit_count = self._feature_delta_sum(x)

        return np.sum(delta_sum, axis=0) / np.sum(hit_count, axis=0) / self.average_path_length(self.samples)

    def apply(self, x):
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        return calc_apply(self.selectors, self.indices, x, num_threads=self.num_threads)

    @classmethod
    def average_path_length(cls, n_nodes):
        """
        Average path length is abstracted because in different cases we may want to
        use a bit different formulas to make the exact match with other software.

        By default we use our own implementation.
        """
        return average_path_length(n_nodes)



from numbers import Real
from typing import Callable

import numpy as np
from scipy.optimize import minimize

from .calc_paths_sum import calc_paths_sum, calc_paths_sum_transpose  # noqa
from .coniferest import Coniferest, ConiferestEvaluator
from .label import Label

__all__ = ["AADForest"]


class AADEvaluator(ConiferestEvaluator):
    def __init__(self, aad):
        super(AADEvaluator, self).__init__(aad, map_value=lambda x: -np.reciprocal(x))
        self.weights = np.full(shape=(self.leaf_count,), fill_value=np.reciprocal(np.sqrt(self.leaf_count)))

    def score_samples(self, x, weights=None):
        """
        Perform the computations.

        Parameters
        ----------
        x
            Features to calculate scores of. Should be C-contiguous for performance.
        weights
            Specific leaf weights to use instead of self.weights

        Returns
        -------
        Array of scores.
        """
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        if weights is None:
            weights = self.weights

        return calc_paths_sum(self.selectors, self.indices, x, weights, num_threads=self.num_threads)

    def loss(
        self,
        weights,
        known_data,
        known_labels,
        anomaly_count,
        nominal_count,
        q_tau,
        C_a=1.0,
        prior_influence=1.0,
        prior_weights=None,
    ):
        """Loss for the known data.

        Adopted from Eq3 of Das et al. 2019 https://arxiv.org/abs/1901.08930
        with the respect to
        1) different anomaly labels - they use +1 for anomalies, -1 for
              nomalies, we use `Label` enum, which is opposite and includes
              `UNKNOWN` label;
        2) different direction of the score axis - they use higher scores
              for anomalies, we use lower scores for anomalies.
        """

        # This new score is negative for the "anomalous" subsample and
        # positive for the "nominal" subsample.
        scores = self.score_samples(known_data, weights) - q_tau

        if prior_weights is None:
            prior_weights = self.weights

        loss = 0.0
        if anomaly_count:
            # For anomalies in "nominal" subsample we add their positive scores.
            loss += C_a * np.sum(scores[(known_labels == Label.ANOMALY) & (scores >= 0)]) / anomaly_count
        if nominal_count:
            # Add for nominals in "anomalous" subsample we add their inverse scores (positive number).
            loss -= np.sum(scores[(known_labels == Label.REGULAR) & (scores <= 0)]) / nominal_count
        delta_weights = weights - prior_weights
        loss += 0.5 * prior_influence * np.inner(delta_weights, delta_weights)

        return loss

    def loss_gradient(
        self,
        weights,
        known_data,
        known_labels,
        anomaly_count,
        nominal_count,
        q_tau,
        C_a=1.0,
        prior_influence=1.0,
        prior_weights=None,
    ):
        scores = self.score_samples(known_data, weights) - q_tau

        if prior_weights is None:
            prior_weights = self.weights

        sample_weights = np.zeros(known_data.shape[0])
        if anomaly_count:
            sample_weights[(known_labels == Label.ANOMALY) & (scores >= 0)] = C_a / anomaly_count
        if nominal_count:
            sample_weights[(known_labels == Label.REGULAR) & (scores <= 0)] = -1.0 / nominal_count

        grad = calc_paths_sum_transpose(
            self.selectors,
            self.indices,
            known_data,
            self.leaf_count,
            sample_weights,
            num_threads=self.num_threads,
        )
        delta_weights = weights - prior_weights
        grad += prior_influence * delta_weights

        return grad

    def loss_hessian(
        self,
        weights,
        vector,
        known_data,
        known_labels,
        q_tau,
        anomaly_count,
        nominal_count,
        C_a=1.0,
        prior_influence=1.0,
        prior_weights=None,
    ):
        return vector * prior_influence


class AADForest(Coniferest):
    """
    Active Anomaly Detection with Isolation Forest.

    See Das et al., 2017 https://arxiv.org/abs/1708.09441

    Parameters
    ----------
    n_trees : int, optional
        Number of trees in the isolation forest.

    n_subsamples : int, optional
        How many subsamples should be used to build every tree.

    max_depth : int or None, optional
        Maximum depth of every tree. If None, `log2(n_subsamples)` is used.

    n_jobs : int or None, optional
        Number of threads to use for scoring. If None - all available CPUs are used.

    random_seed : int or None, optional
        Random seed to use for reproducibility. If None - random seed is used.

    prior_influence : float or callable, optional
        An regularization coefficient value in the loss functioin. Default is 1.0.
        Signature: '(anomaly_count, nominal_count) -> float'
    """

    def __init__(
        self,
        n_trees=100,
        n_subsamples=256,
        max_depth=None,
        tau=0.97,
        C_a=1.0,
        prior_influence=1.0,
        n_jobs=None,
        random_seed=None,
    ):
        super().__init__(
            trees=[],
            n_subsamples=n_subsamples,
            max_depth=max_depth,
            n_jobs=n_jobs,
            random_seed=random_seed,
        )
        self.n_trees = n_trees
        self.tau = tau
        self.C_a = C_a

        if isinstance(prior_influence, Callable):
            self.prior_influence = prior_influence
        elif isinstance(prior_influence, Real):
            self.prior_influence = lambda anomaly_count, nominal_count: prior_influence
        else:
            raise ValueError("prior_influence is neither a callable nor a constant")

        self.evaluator = None

    def _build_trees(self, data):
        if len(self.trees) == 0:
            self.trees = self.build_trees(data, self.n_trees)
            self.evaluator = AADEvaluator(self)

    def fit(self, data, labels=None):
        """
        Build the trees with the data `data`.

        Parameters
        ----------
        data
            Array with feature values of objects.

        labels
            Optional. Labels of objects. May be regular, anomalous or unknown.
            See `Label` data for details.

        Returns
        -------
        self
        """
        # If no labels were supplied, train with them.
        if labels is None:
            return self.fit_known(data)

        # Otherwise select known data, and train on it.
        labels = np.asarray(labels)
        index = labels != Label.UNKNOWN
        return self.fit_known(data, data[index, :], labels[index])

    def fit_known(self, data, known_data=None, known_labels=None):
        """
        The same `fit` but with a bit of different API. Known data and labels
        are separated from training data for time and space optimality. High
        chances are that `known_data` is much smaller that `data`. At that case
        it is not reasonable to hold the labels for whole `data`.

        Parameters
        ----------
        data
            Training data (array with feature values) to build trees with.

        known_data
            Feature values of known data.

        known_labels
            Labels of known data.

        Returns
        -------
        self
        """

        known_data = np.asarray(known_data) if known_data is not None else None
        known_labels = np.asarray(known_labels) if known_labels is not None else None

        self._build_trees(data)

        if (
            known_data is None
            or len(known_data) == 0
            or known_labels is None
            or len(known_labels) == 0
            or np.all(known_labels == Label.UNKNOWN)
        ):
            return self

        scores = self.score_samples(data)
        # Our scores are negative, so we need to "invert" the quantile.
        q_tau = np.quantile(scores, 1.0 - self.tau)

        anomaly_count = np.count_nonzero(known_labels == Label.ANOMALY)
        nominal_count = np.count_nonzero(known_labels == Label.REGULAR)
        prior_influence = self.prior_influence(anomaly_count, nominal_count)

        def fun(weights):
            return self.evaluator.loss(
                weights,
                known_data,
                known_labels,
                anomaly_count,
                nominal_count,
                q_tau,
                self.C_a,
                prior_influence,
            )

        def jac(weights):
            return self.evaluator.loss_gradient(
                weights,
                known_data,
                known_labels,
                anomaly_count,
                nominal_count,
                q_tau,
                self.C_a,
                prior_influence,
            )

        def hessp(weights, vector):
            return self.evaluator.loss_hessian(
                weights,
                vector,
                known_data,
                known_labels,
                anomaly_count,
                nominal_count,
                q_tau,
                self.C_a,
                prior_influence,
            )

        res = minimize(
            fun,
            self.evaluator.weights,
            method="trust-krylov",
            jac=jac,
            hessp=hessp,
            tol=1e-4,
        )
        weights_norm = np.sqrt(np.inner(res.x, res.x))
        self.evaluator.weights = res.x / weights_norm

        return self

    def score_samples(self, samples):
        """
        Computer scores for the supplied data.

        Parameters
        ----------
        samples
            Feature values to compute scores on.

        Returns
        -------
        Array with computed scores.
        """
        return self.evaluator.score_samples(samples)

    def feature_signature(self, x):
        raise NotImplementedError()

    def feature_importance(self, x):
        raise NotImplementedError()

    def apply(self, x):
        """
        Apply the forest to X, return leaf indices.

        Parameters
        ----------
        x : ndarray shape (n_samples, n_features)
            2-d array with features.

        Returns
        -------
        x_leafs : ndarray of shape (n_samples, n_estimators)
            For each datapoint x in X and for each tree in the forest,
            return the index of the leaf x ends up in.
        """
        return self.evaluator.apply(x)



import os
import sys

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

extra_compile_args = []
extra_link_args = []
# macOS Clang doesn't support OpenMP, but we allow to force build with it anyway
if sys.platform != "darwin" or os.environ.get("CONIFEREST_FORCE_OPENMP_ON_MACOS", False):
    extra_compile_args.append("-fopenmp")
    extra_link_args.append("-fopenmp")


extensions = [
    Extension(
        "coniferest.calc_paths_sum",
        ["src/coniferest/calc_paths_sum.pyx"],
        include_dirs=[np.get_include()],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]


setup(
    ext_modules=cythonize(extensions),
    cmdclass={"build_ext": build_ext},
)



import numpy as np
from cython.parallel cimport prange, parallel

cimport numpy as np
cimport cython


def calc_paths_sum(selector_t [::1] selectors,
                   np.int64_t [::1] indices,
                   floating [:, ::1] data,
                   np.float64_t [::1] weights=None,
                   int num_threads=1):
    cdef np.ndarray [np.double_t, ndim=1] paths = np.zeros(data.shape[0])
    cdef np.float64_t [::1] paths_view = paths
    cdef Py_ssize_t sellen = selectors.shape[0]

    if np.any(np.diff(indices) < 0):
        raise ValueError('indices should be an increasing sequence')

    if indices[-1] > sellen:
        raise ValueError('indices are out of range of the selectors')

    _paths_sum(selectors, indices, data, paths_view, weights, num_threads)
    return paths


def calc_paths_sum_transpose(selector_t [::1] selectors,
                             np.int64_t [::1] indices,
                             floating [:, ::1] data,
                             Py_ssize_t leaf_count,
                             np.float64_t [::1] weights=None,
                             int num_threads=1):
    cdef np.ndarray [np.double_t, ndim=1] values = np.zeros(leaf_count)
    cdef np.float64_t [::1] values_view = values
    cdef Py_ssize_t sellen = selectors.shape[0]

    if np.any(np.diff(indices) < 0):
        raise ValueError('indices should be an increasing sequence')

    if indices[-1] > sellen:
        raise ValueError('indices are out of range of the selectors')

    if weights is not None and weights.shape[0] != data.shape[0]:
        raise ValueError('data and weights should have the same length')

    _paths_sum_transpose(selectors, indices, data, values_view, weights, num_threads)
    return values


def calc_feature_delta_sum(selector_t [::1] selectors,
                   np.int64_t [::1] indices,
                   floating [:, ::1] data,
                   int num_threads=1):
    cdef np.ndarray [np.double_t, ndim=2] delta_sum = np.zeros([data.shape[0], data.shape[1]])
    cdef np.float64_t [:, ::1] delta_sum_view = delta_sum
    cdef np.ndarray [np.int64_t, ndim=2] hit_count = np.zeros([data.shape[0], data.shape[1]], dtype=np.int64)
    cdef np.int64_t [:, ::1] hit_count_view = hit_count
    cdef Py_ssize_t sellen = selectors.shape[0]

    if np.any(np.diff(indices) < 0):
        raise ValueError('indices should be an increasing sequence')

    if indices[-1] > sellen:
        raise ValueError('indices are out of range of the selectors')

    _feature_delta_sum(selectors, indices, data, delta_sum_view, hit_count_view, num_threads)
    return delta_sum, hit_count

def calc_apply(selector_t [::1] selectors, np.int64_t [::1] indices, floating [:, ::1] data, int num_threads=1):
    cdef np.ndarray [np.int64_t, ndim=2] leafs = np.zeros([data.shape[0], indices.shape[0] - 1], dtype=np.int64)
    cdef np.int64_t [:, ::1] leafs_view = leafs
    cdef Py_ssize_t sellen = selectors.shape[0]

    if np.any(np.diff(indices) < 0):
        raise ValueError('indices should be an increasing sequence')

    if indices[-1] > sellen:
        raise ValueError('indices are out of range of the selectors')

    _apply(selectors, indices, data, leafs_view, num_threads)
    return leafs



@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _paths_sum(selector_t [::1] selectors,
                     np.int64_t [::1] indices,
                     floating [:, ::1] data,
                     np.float64_t [::1] paths,
                     np.float64_t [::1] weights=None,
                     int num_threads=1):

    cdef Py_ssize_t trees
    cdef Py_ssize_t tree_index
    cdef Py_ssize_t x_index
    cdef selector_t selector
    cdef Py_ssize_t tree_offset
    cdef np.int32_t feature, i

    with nogil, parallel(num_threads=num_threads):
        trees = indices.shape[0] - 1

        for x_index in prange(data.shape[0], schedule='static'):
            for tree_index in range(trees):
                tree_offset = indices[tree_index]
                i = 0
                while True:
                    selector = selectors[tree_offset + i]
                    feature = selector.feature
                    if feature < 0:
                        break

                    if data[x_index, feature] <= selector.value:
                        i = selector.left
                    else:
                        i = selector.right

                if weights is None:
                    paths[x_index] += selector.value
                else:
                    paths[x_index] += weights[selector.left] * selector.value


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _paths_sum_transpose(selector_t [::1] selectors,
                               np.int64_t [::1] indices,
                               floating [:, ::1] data,
                               np.float64_t [::1] values,
                               np.float64_t [::1] weights=None,
                               int num_threads=1):

    cdef Py_ssize_t trees
    cdef Py_ssize_t tree_index
    cdef Py_ssize_t x_index
    cdef selector_t selector
    cdef Py_ssize_t tree_offset
    cdef np.int32_t feature, i

    with nogil, parallel(num_threads=num_threads):
        trees = indices.shape[0] - 1

        for tree_index in prange(trees, schedule='static'):
            tree_offset = indices[tree_index]
            for x_index in range(data.shape[0]):
                i = 0
                while True:
                    selector = selectors[tree_offset + i]
                    feature = selector.feature
                    if feature < 0:
                        break

                    if data[x_index, feature] <= selector.value:
                        i = selector.left
                    else:
                        i = selector.right

                if weights is None:
                    values[selector.left] += selector.value
                else:
                    values[selector.left] += weights[x_index] * selector.value


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _feature_delta_sum(selector_t [::1] selectors,
                     np.int64_t [::1] indices,
                     floating [:, ::1] data,
                     np.float64_t [:, ::1] delta_sum,
                     np.int64_t [:, ::1] hit_count,
                     int num_threads=1):

    cdef Py_ssize_t trees
    cdef Py_ssize_t tree_index
    cdef Py_ssize_t x_index
    cdef selector_t selector, child_selector
    cdef Py_ssize_t tree_offset
    cdef np.int32_t feature, i

    with nogil, parallel(num_threads=num_threads):
        trees = indices.shape[0] - 1

        for x_index in prange(data.shape[0], schedule='static'):
            for tree_index in range(trees):
                tree_offset = indices[tree_index]
                i = 0
                while True:
                    selector = selectors[tree_offset + i]
                    feature = selector.feature
                    if feature < 0:
                        break

                    if data[x_index, feature] <= selector.value:
                        i = selector.left
                    else:
                        i = selector.right

                    child_selector = selectors[tree_offset + i]

                    delta_sum[x_index, feature] += 1.0 + 2.0 * (child_selector.log_n_node_samples - selector.log_n_node_samples)
                    hit_count[x_index, feature] += 1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _apply(selector_t [::1] selectors,
                 np.int64_t [::1] indices,
                 floating [:, ::1] data,
                 np.int64_t [:, ::1] leafs,
                 int num_threads=1):

    cdef Py_ssize_t trees
    cdef Py_ssize_t tree_index
    cdef Py_ssize_t x_index
    cdef selector_t selector
    cdef Py_ssize_t tree_offset
    cdef np.int32_t feature, i

    with nogil, parallel(num_threads=num_threads):
        trees = indices.shape[0] - 1

        for x_index in prange(data.shape[0], schedule='static'):
            for tree_index in range(trees):
                tree_offset = indices[tree_index]
                i = 0
                while True:
                    selector = selectors[tree_offset + i]
                    feature = selector.feature
                    if feature < 0:
                        break

                    if data[x_index, feature] <= selector.value:
                        i = selector.left
                    else:
                        i = selector.right

                leafs[x_index, tree_index] = selector.left
